#|
This file is part of Ordo.

Copyright (C) 2025 Ray Miller

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, version 3.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <https://www.gnu.org/licenses/>.
|#

(define-module (ordo blueprint)
  #:use-module (ice-9 exceptions)
  #:use-module (ice-9 format)
  #:use-module (oop goops)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-235)
  #:export (<task>
            task
            task?
            task-name
            task-pre-condititon
            task-action
            task-args
            task-trigger

            <handler>
            handler
            handler?
            handler-name
            handler-action
            handler-args

            <blueprint>
            blueprint
            blueprint?
            blueprint-name
            blueprint-steps
            blueprint-handlers))

(define-class <task> ()
  (name #:init-keyword #:name #:getter task-name)
  (pre-condition #:init-keyword #:pre-condition #:getter task-pre-condition)
  (action #:init-keyword #:action #:getter task-action)
  (args #:init-keyword #:args #:getter task-args)
  (trigger #:init-keyword #:trigger #:getter task-trigger))

(define (task . args) (apply make <task> args))
(define (task? x) (is-a? x <task>))

(define-class <handler> ()
  (name #:init-keyword #:name #:getter handler-name)
  (action #:init-keyword #:action #:getter handler-action)
  (args #:init-keyword #:args #:getter handler-args))

(define (handler . args) (apply make <handler> args))
(define (handler? x) (is-a? x <handler>))

(define-class <blueprint> ()
  (name #:init-keyword #:name #:getter blueprint-name)
  (steps #:init-keyword #:steps #:getter blueprint-steps)
  (handlers #:init-keyword #:handlers #:getter blueprint-handlers))

(define (blueprint? x) (is-a? x <blueprint>))

(define (validate-triggers blueprint-name tasks handlers)
  (let ((handler-names (map handler-name handlers)))
    (for-each (lambda (task)
                (for-each (lambda (trigger)
                            (unless (member trigger handler-names)
                              (raise-exception
                               (make-exception
                                (make-programming-error)
                                (make-exception-with-message (format #f "Task ~a in blueprint ~a references unknown trigger: ~a"
                                                                     blueprint-name (task-name task) trigger))))))
                          (task-trigger task)))
              tasks)))

(define (blueprint name . args)
  (let ((steps (filter (disjoin task? blueprint?) args))
        (handlers (filter handler? args)))
    (validate-triggers name (filter task? steps) handlers)
    (make <blueprint> #:name name #:steps steps #:handlers handlers)))
